/*
 * Copyright (c) 2007, 2009-2010 Tamas Tevesz <ice@extreme.hu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/types.h>

#include <bsd_auth.h>
#include <err.h>
#include <grp.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <utmp.h>		/* UT_NAMESIZE */

#ifndef	VPN_GROUP
#define	VPN_GROUP		"_openvpnusers"
#endif

/*
 * Sync this with OpenVPN's misc.h:USER_PASS_LEN
 */
#define	OPENVPN_USER_PASS_MAXLEN	(127 + 1)

#define	OPENVPN_AUTH_OK			0
#define	OPENVPN_AUTH_FAIL		1

extern	char	*__progname;
	int	 is_ingroup(char *);
	void	 logx(int, int, const char *, ...)
		    __attribute__((__noreturn__, format(printf, 3, 4)));

	char	 username[UT_NAMESIZE+1];
	char	 password[OPENVPN_USER_PASS_MAXLEN];

int main(int argc, char **argv) {
	char *tmp;

	(void)memset(username, 0, sizeof(username));
	(void)memset(password, 0, sizeof(password));

	if (argc == 2) {
		FILE *fp;
		if ((fp = fopen(argv[1], "r")) == NULL)
			logx(OPENVPN_AUTH_FAIL, LOG_ERR,
			    "Can not open password file \"%s\"", argv[1]);

		if ((tmp = fgets(username, sizeof(username), fp)) == NULL)
			logx(OPENVPN_AUTH_FAIL, LOG_ERR,
			    "Unable to read username from \"%s\"", argv[1]);

		if ((tmp = fgets(password, sizeof(password), fp)) == NULL)
			logx(OPENVPN_AUTH_FAIL, LOG_ERR,
			    "Unable to read password from \"%s\"", argv[1]);

		(void)fclose(fp);
		username[(long)strcspn(username, "\r\n")] = '\0';
		password[(long)strcspn(password, "\r\n")] = '\0';
	} else {
		if ((tmp = getenv("username")) == NULL)
			logx(OPENVPN_AUTH_FAIL, LOG_ERR,
			    "No \"username\" found in the environment");

		/* should this happen, username max size needs to be revisited */
		if (strlcpy(username, tmp, sizeof(username)) >= sizeof(username))
			logx(OPENVPN_AUTH_FAIL, LOG_ERR,
			    "Internal error (username buffer overflow)");

		if ((tmp = getenv("password")) == NULL)
			logx(OPENVPN_AUTH_FAIL, LOG_ERR,
			    "No \"password\" found in the environment");

		/* should this happen, password max size needs to be revisited */
		if (strlcpy(password, tmp, sizeof(password)) >= sizeof(password))
			logx(OPENVPN_AUTH_FAIL, LOG_ERR,
			    "Internal error (password buffer overflow)");
	}

	if (!is_ingroup(username))
		logx(OPENVPN_AUTH_FAIL, LOG_ERR,
		    "User %s is not a member of " VPN_GROUP , username);

	if (auth_userokay(username, NULL, NULL, password) != 0)
		logx(OPENVPN_AUTH_OK, LOG_INFO,
		    "Accepted password for %s", username);

	logx(OPENVPN_AUTH_FAIL, LOG_ERR, "Bad password for %s", username);

	/* NOTREACHED */
	return OPENVPN_AUTH_FAIL;
}

/* butchered from src/usr.bin/id/id.c */
int is_ingroup(char *user) {
	gid_t groups[NGROUPS+1], vpngroup_gid;
	struct group *gr;
	struct passwd *pw;
	int ngroups, i;
	
	if ((gr = getgrnam(VPN_GROUP)) == NULL)
		logx(OPENVPN_AUTH_FAIL, LOG_ERR,
		    "Can not find group %s in the system group database", VPN_GROUP);

	vpngroup_gid = gr->gr_gid;

	if ((pw = getpwnam(user)) == NULL)
		logx(OPENVPN_AUTH_FAIL, LOG_ERR, "Bad user %s", user);

	ngroups = NGROUPS + 1;
	if (getgrouplist(pw->pw_name, pw->pw_gid, groups, &ngroups) == -1)
		logx(OPENVPN_AUTH_FAIL, LOG_ERR,
		    "System group database is corrupt "
		    "(user is in more than %d groups)", NGROUPS);

	for (i = 0; i < ngroups; i++)
		if (vpngroup_gid == groups[i])
			return 1;

	return 0;
}

void logx(int exitval, int prio, const char *fmt, ...) {
	va_list va;

	va_start(va, fmt);
	openlog(__progname, LOG_PID|LOG_NDELAY, LOG_AUTH);
	vsyslog(prio, fmt, va);
	closelog();
	va_end(va);

	(void)memset(username, 0, sizeof(username));
	(void)memset(password, 0, sizeof(password));

	exit(exitval);
}
